from pangeamt_nlp.translation_model.translation_model_factory import (
    TranslationModelFactory,
)
import logging
from pangeamt_nlp.seg import Seg
from nteu_corporate_engine.pipeline import Pipeline
from typing import Dict, List
from asyncio import Lock
import math
import itertools
import numpy as np

NO_LONELY_TRANSLATABLE_CHARS = "\n\t/n/t \u200b"

class Engine:
    def __init__(self, config: Dict, log_file: str = None, dbug: bool = False):
        self.setup_log(log_file, dbug)
        self._config = config
        self._pipeline = Pipeline(self._config)
        self._model = self.load_model()
        online_learning_args = config["translation_engine_server"].get(
            "online_learning", None
        )
        if online_learning_args is not None:
            self._model.setup_online_trainer(online_learning_args)

    def load_model(self):
        name = self._config["translation_model"]["name"]
        args = self._config["translation_model"]["args_decoding"]
        if self._config["translation_engine_server"]["gpu"]:
            args["gpu"] = 0
        model_path = self._config["translation_engine_server"]["model_path"]
        if args == {}:
            msg = f"Loading Model -> {name} with default args"
        else:
            msg = f"Loading Model -> {name} with arguments {args}."
        logging.info(msg)
        translation_model = TranslationModelFactory.get_class(name)
        return translation_model(model_path, **args)

    async def translate(self, srcs: List, n_best: int = 1):
        best_translations = []
        best_scores = []
        if self._config["translation_model"]["name"] == "onmt":
            translations = self._model.translate(srcs, n_best=n_best)
            for translation in translations:
                self.log_translation(translation)
                best_translations.append(translation.pred_sents[0].sentence)
                best_scores.append(translation.pred_sents[0].score)

        elif self._config["translation_model"]["name"] == "marianmt":
            args_dec = self._config["translation_model"]["args_decoding"]
            if not args_dec["in_training_servers"]:
                best_translations = self._model.translate(srcs, n_best=n_best)
            else:
                model_path = self._config["translation_engine_server"]["model_path"]
                marian_path = self._config["translation_model"]["marian_path"]
                if self._config["translation_engine_server"]["gpu"]:
                    gpu = 0
                else:
                    gpu = None
                best_translations = self._model.translate_training_servers(
                    srcs, model_path, marian_path, gpu
                )
            for translation in best_translations:
                self.log_translation(translation)
        return best_translations,best_scores


    async def split_sentence(self, seg_src: str, max_len: int):

        chunked_input = []
        rest_seg = seg_src
        separators = [".","?","!",":",";"]
        while len(rest_seg) > max_len:
            good_str = rest_seg[:max_len]
            last_separators = []
            for separator in separators:
                last_separators.append(good_str.rfind(separator))
        
            last_sep = max(last_separators)
            if last_sep == -1:
                last_sep = good_str.rfind(" ")
            
            if last_sep == -1:
                last_sep = max_len
            
            rest_seg = rest_seg[last_sep+1:].strip()
            final_segment = good_str[:last_sep+1].strip()

            seg = Seg(final_segment)
            await self._pipeline.preprocess(seg)
            chunked_input.append(seg.src)

        if rest_seg.strip(NO_LONELY_TRANSLATABLE_CHARS) != "":
            seg = Seg(rest_seg)
            await self._pipeline.preprocess(seg)
            chunked_input.append(seg.src)

        return chunked_input


    async def process_batch(self, batch: List, lock: Lock = None):
        srcs = []
        segs = []
        ans = []
        
        no_trans_ind = []
        no_trans_score = []
        no_trans_trad = []

        max_len = 600
        
        for i, src in enumerate(batch):
            
            seg = Seg(src)

            # Model can not translate only spaces, new lines, tabs...
            if src.strip(NO_LONELY_TRANSLATABLE_CHARS) == "":
                no_trans_ind.append(i)
                no_trans_score.append(0)
                no_trans_trad.append(seg.src)
            
            # Model does not translate properly son long segments
            # So we split long segments and translate them before 
            elif len(seg.src) > max_len:
                
                logging.info("Splitting too long segment: "+ str(seg.src))
                logging.info("With len: "+ str(len(seg.src)))
                
                seg.src = seg.src.strip()
                chunked_segment = await self.split_sentence(seg.src, max_len)
                if lock is not None:
                    async with lock:
                        translations,scores = await self.translate(chunked_segment)
                else:
                    translations,scores = await self.translate(chunked_segment)
                
                trans = " ".join(translations)
                score = sum(scores) / len(scores)
                
                no_trans_ind.append(i)
                no_trans_trad.append(trans)
                no_trans_score.append(score)

                await self._pipeline.preprocess(seg)

            else:
                await self._pipeline.preprocess(seg)
                srcs.append(seg.src)

            segs.append(seg)


        # If there is nothing to translate skip translation
        if len(no_trans_ind) != len(batch):

            # Only for GPU
            if self._config["translation_engine_server"]["gpu"]:
                biggest = len(max(srcs, key=len))
                # change batch size depending on max sentence length to not overload gpu
                batch_size = 32
                logging.info("MAX CHARACTERS SENTENCE: " + str(biggest))
                logging.info("LEN BATCH TO TRANSLATE: " + str(len(srcs)))
                logging.info("BATCH SIZE USED: " + str(batch_size))

                if batch_size >= len(srcs):
                    if lock is not None:
                        async with lock:
                            translations,scores = await self.translate(srcs)
                    else:
                        translations,scores = await self.translate(srcs)

                else:
                    translations = []
                    scores = []
                    iters = math.ceil(len(srcs)/batch_size)
                    for i in range(iters):
                        if lock is not None:
                            async with lock:
                                trans,scs = await self.translate(srcs[i*batch_size:(i+1)*batch_size])
                        else:
                            trans,scs = await self.translate(srcs[i*batch_size:(i+1)*batch_size])

                        translations = translations + trans
                        scores = scores + scs

            # Only for CPU
            else:
                if lock is not None:
                    async with lock:
                        translations,scores = await self.translate(srcs)
                else:
                    translations,scores = await self.translate(srcs)

        else:                      
            logging.info("Nothing to translate in "+ str(batch))
            translations = []  
            scores = []

        # We add back eliminated alements
        for e, score, trad in zip(no_trans_ind, no_trans_score, no_trans_trad):
            translations.insert(e, trad)
            scores.insert(e, score)
        
        final_scores = [np.e**(s)*10 for s in scores]

        for translation, seg in zip(translations, segs):
            seg.tgt_raw = translation
            seg.tgt = seg.tgt_raw
            if seg.tgt_raw.strip(NO_LONELY_TRANSLATABLE_CHARS) != "":
                await self._pipeline.postprocess(seg)
            ans.append(seg.tgt)
            logging.info(
                f"Translated -> {seg.src_raw} -> {seg.src} "
                f"-> {seg.tgt_raw} -> {seg.tgt}"
            )
        return ans, final_scores

    async def online_train(self, batch: List, num_steps: int = 1, lock: Lock = None):
        tus = []
        for src, tgt in [tu for tu in batch]:
            seg = Seg(src, tgt)
            await self._pipeline.process_train(seg)
            tus.append((seg.src, seg.tgt))
        if lock is not None:
            async with lock:
                self._model.online_train(tus, num_steps)
        else:
            self._model.online_train(tus, num_steps)

    @classmethod
    def setup_log(cls, log_file: str = None, dbug: bool = None):
        hdlrs = [logging.StreamHandler()]
        if log_file is not None:
            hdlrs.append(logging.FileHandler(log_file))
        cls.lvl = logging.DEBUG if dbug else logging.INFO
        logging.basicConfig(
            handlers=hdlrs,
            level=cls.lvl,
            format="%(asctime)s :: %(levelname)s :: %(message)s",
        )

    def log_translation(self, translation: "Translation"):
        if self.lvl == logging.DEBUG:
            n_best = len(translation)
            log_msg = (
                f"For sentence {translation.pred_sents[0].src_raw}",
                f"{n_best}-best translations:",
            )
            logging.debug(log_msg)
            f = "\nOption #{0} with score {1}:\n{2}\nAttention:\n"
            for i, prediction in enumerate(translation.pred_sents):
                log_msg = f.format(i, prediction.score, prediction.sentence)
                table = prediction.get_pretty_attention()
                if table is not None:
                    table_fields = table._field_names
                    table_width = len(table_fields)
                    for column_index in range(0, table_width, 3):
                        f_index = column_index
                        l_index = min([column_index + 3, table_width])
                        fields_to_take = table_fields[f_index:l_index]
                        str_to_add = table.get_string(fields=fields_to_take) + "\n"
                        log_msg += str_to_add
                logging.debug(log_msg)



from pangeamt_nlp.processor.pipeline_decoding import PipelineDecoding
from pangeamt_nlp.processor.pipeline_training import PipelineTraining
from pangeamt_nlp.truecaser.truecaser import Truecaser
from pangeamt_nlp.bpe.bpe import BPE
from pangeamt_nlp.bpe.sentencepiece import SentencePieceSegmenter
from pangeamt_nlp.tokenizer.tokenizer_factory import TokenizerFactory
from pangeamt_nlp.tokenizer.tokenizer_base import TokenizerBase
from pangeamt_nlp.seg import Seg
from typing import Dict, Tuple
import os
import logging
from logging import Logger


class Pipeline:

    def __init__(self, config: Dict) -> "Pipeline":
        self._config = config
        self._src_lang = config["src_lang"]
        self._tgt_lang = config["tgt_lang"]
        self._decoding_pipeline = self.load_decoding_pipeline()
        self._training_pipeline = self.load_training_pipeline()
        self._src_bpe, self._tgt_bpe = self.load_bpe()
        self._src_truecaser, self._tgt_truecaser = self.load_truecaser()
        self._src_tokenizer, self._tgt_tokenizer = self.load_tokenizer()

    async def preprocess(self, seg: Seg):
        self._decoding_pipeline.process_src(seg, logger=logging.getLogger())
        seg.src = self._src_tokenizer.tokenize(seg.src)
        logging.debug("Tokenizing: " + seg.src)
        if self._src_truecaser is not None:
            seg.src = self._src_truecaser.truecase(seg.src)
            logging.debug("Truecasing: " + seg.src)
        if self._src_bpe is not None:
            seg.src = self._src_bpe.apply(seg.src)
            logging.debug("BPE: " + seg.src)

    async def postprocess(self, seg: Seg):
        seg.tgt = self._tgt_bpe.undo(seg.tgt)
        logging.debug("Undo BPE" + seg.tgt)
        seg.tgt = self._tgt_tokenizer.detokenize(seg.tgt.split(" "))
        logging.debug("Detokenize: " + seg.tgt)
        if self._tgt_truecaser:
            seg.tgt = self._tgt_truecaser.detruecase(seg.tgt)
            logging.debug("Detruecasing: " + seg.tgt)
        self._decoding_pipeline.process_tgt(seg, logger=logging.getLogger())

    async def process_train(self, seg: Seg):
        self._training_pipeline.normalize(seg)
        seg.src = self._src_tokenizer.tokenize(seg.src)
        seg.tgt = self._tgt_tokenizer.tokenize(seg.tgt)
        logging.debug("Tokenizing: " + seg.src + ", " + seg.tgt)
        if self._src_truecaser is not None:
            seg.src = self._src_truecaser.truecase(seg.src)
            logging.debug("Truecasing: " + seg.src)
        if self._tgt_truecaser is not None:
            seg.tgt = self._tgt_truecaser.truecase(seg.tgt)
            logging.debug("Truecasing: " + seg.tgt)
        if self._src_bpe is not None:
            seg.src = self._src_bpe.apply(seg.src)
            logging.debug("BPE: " + seg.src)
        if self._tgt_bpe is not None:
            seg.tgt = self._tgt_bpe.apply(seg.tgt)
            logging.debug("BPE: " + seg.tgt)

    def load_decoding_pipeline(self) -> PipelineDecoding:
        return PipelineDecoding.create_from_dict(
            self._src_lang, self._tgt_lang, self._config["processors"]
        )

    def load_training_pipeline(self) -> PipelineTraining:
        return PipelineTraining.create_from_dict(
            self._src_lang, self._tgt_lang, self._config["processors"]
        )

    def load_bpe(self) -> Tuple[BPE, BPE]:
        if self._config["bpe"] is not None:
            path = self._config["translation_engine_server"]["bpe"]
            if self._config["bpe"].get("sentencepiece"):
                src_bpe, tgt_bpe = self.load_sentencepiece(path)
            else:
                if self._config["bpe"]["joint"]:
                    src_bpe, tgt_bpe = self.load_joint_bpe(path)
                else:
                    src_bpe, tgt_bpe = self.load_split_bpe(path)
            return src_bpe, tgt_bpe
        else:
            return None, None

    def load_sentencepiece(self, path: str):
        src_model = os.path.join(path, "src_codes.model")
        tgt_model = os.path.join(path, "tgt_codes.model")
        return (
            SentencePieceSegmenter(src_model),
            SentencePieceSegmenter(tgt_model)
        )

    def load_joint_bpe(self, path: str) -> Tuple[BPE, BPE]:
        codes = os.path.join(path, "codes32k.txt")
        src_vocab = os.path.join(path, "src_vocab.txt")
        tgt_vocab = os.path.join(path, "tgt_vocab.txt")
        return BPE(codes, src_vocab), BPE(codes, tgt_vocab)

    def load_split_bpe(self, path: str) -> Tuple[BPE, BPE]:
        src_codes = os.path.join(path, "src_codes.txt")
        tgt_codes = os.path.join(path, "tgt_codes.txt")
        return BPE(src_codes), BPE(tgt_codes)

    def load_truecaser(self) -> Tuple[Truecaser, Truecaser]:
        src_truecaser, tgt_truecaser = None, None
        if self._config["truecaser"]["src"] == "enabled":
            path = os.path.join(
                self._config["translation_engine_server"]["truecaser"],
                "src_model.txt"
            )
            src_truecaser = Truecaser(path)
        if self._config["truecaser"]["tgt"] == "enabled":
            path = os.path.join(
                self._config["translation_engine_server"]["truecaser"],
                "tgt_model.txt"
            )
            tgt_truecaser = Truecaser(path)
        return src_truecaser, tgt_truecaser

    def load_tokenizer(self) -> Tuple[TokenizerBase, TokenizerBase]:
        src_tok_name = self._config["tokenizer"]["src"]
        tgt_tok_name = self._config["tokenizer"]["tgt"]

        src_tokenizer = TokenizerFactory.new(self._src_lang, src_tok_name)
        tgt_tokenizer = TokenizerFactory.new(self._tgt_lang, tgt_tok_name)

        return (src_tokenizer, tgt_tokenizer)



#!/usr/bin/env python3.7
import os
import sys
import yaml
import logging
import uvloop
import asyncio
import pickle
from nteu_corporate_engine.engine import Engine
from aiohttp import web
from typing import Dict
import numpy as np

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


class TranslationEngineServer:

    def __init__(self, config: str = None, serialized: bool = False):
        self._app = web.Application(client_max_size=1024**3)
        self._app["lock"] = asyncio.Lock()

        if serialized:
            if os.environ.get("PYTHONTRANSLATIONSERVERDEBUG", None) is not None:
                dbug = True 
            else:
                dbug = False
            Engine.setup_log(dbug=dbug)
            with open("engine.pkl", "rb") as engine_bts:
                self._app["engine"] = pickle.load(engine_bts)
        else:
            with open(config, "r") as file:
                config = yaml.safe_load(file)
            self._update_config(config)
            self._app["engine"] = Engine(config)
        self._app.router.add_post("/translate", handle_translate)
        self._app.router.add_post("/train", handle_train)
        self._app.router.add_post("/isready", handle_ready)

        self._app.router.add_get("/healthcheck", handle_healthcheck)

    def start(self, port: int):
        web.run_app(self._app, port=port)

    def _update_config(self, root: Dict, prefix: str = None):
        for key in root.keys():
            name = prefix + key.upper() if prefix is not None else key.upper()
            value = root[key]
            if isinstance(value, dict):
                self._update_config(value, name + "_")
            elif os.environ.get(name) is not None:
                root[key] = os.environ.get(name)


async def handle_translate(req):
    lock = req.app["lock"]
    engine = req.app["engine"]

    try:
        req = await req.json()
        batch = None
        batch = req.get("srcs")
        translations = None
        translations, scores = await engine.process_batch(batch, lock)
        ans = {
            "tus": []
        }
        for src, translation, score in zip(batch, translations, scores):
            ans["tus"].append({"src": src, "tgt": translation, "score": score})
        return web.json_response(ans, status=200)

    except Exception as e:
        logging.exception("ERROR in handle_translate: " + str(e))
        logging.info("BATCH: " + str(batch))
        logging.info("TRANSLATIONS: " + str(translations))
        logging.info("REQ: " + str(req))
        response_obj = {'status': 'failed', 'reason': str(e)}
        return web.json_response(response_obj, status=500)


async def handle_train(req):
    lock = req.app["lock"]
    engine = req.app["engine"]

    try:
        req = await req.json()
        num_steps = req.get("steps", 2)
        translation_units = req.get("tus")
        batch = []
        for tu in translation_units:
            if tu["tgt"] == "":
                raise Exception(
                    "Missing target segment for Online Learning Training."
                )
            batch.append((tu["src"], tu["tgt"]))
        await engine.online_train(batch, num_steps, lock)
        return web.json_response({"rc": 0}, status=200)
    except Exception as e:
        logging.exception(str(e))
        response_obj = {'status': 'failed', 'reason': str(e)}
        return web.json_response(response_obj, status=500)


async def handle_ready(req):
    return web.json_response({"rc": 0}, status=200)


async def handle_healthcheck(req):
    return web.json_response({"OK": "OK"}, status=200)


if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) < 2:
        print("Three arguments needed: ")
        print("config and port")
    else:
        try:
            server = TranslationEngineServer(args[0])
            server.start(int(args[1]))
        except Exception as e:
            logging.exception(str(e))
            raise Exception("Error") from e
